# Module of TWiki Enterprise Collaboration Platform, http://TWiki.org/
#
# Copyright (C) 1999-2007 Peter Thoeny, peter@thoeny.org
# and TWiki Contributors. All Rights Reserved. TWiki Contributors
# are listed in the AUTHORS file in the root of this distribution.
# NOTE: Please extend that file, not this notice.
#
# Additional copyrights apply to some or all of the code in this
# file as follows:
# Based on parts of Ward Cunninghams original Wiki and JosWiki.
# Copyright (C) 1998 Markus Peter - SPiN GmbH (warpi@spin.de)
# Some changes by Dave Harris (drh@bhresearch.co.uk) incorporated
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version. For
# more details read LICENSE in the root of this distribution.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# As per the GPL, removal of this notice is prohibited.
#
# AdvancedManageContrib is developed by 
# Sascha Vetter (cc/sve) 2008

=pod

---+ package TWiki::Contrib::AdvancedManageContrib

Contrib function to move topics and set parent

=cut

package TWiki::Contrib::AdvancedManageContrib;

use strict;
use Error qw( :try );
use Assert;
use CGI::Carp qw(fatalsToBrowser);

require TWiki;
require TWiki::UI;
require TWiki::UI::Manage;

my $mydebug = 0;

=pod

---++ StaticMethod move($session)

Command handler for =move= command.

=cut

sub move {
    my $session = shift;
    my $store = $session->{store};
    my $this = bless({});
    my $query = $session->{cgiQuery};
    
    $this->{fromWeb} = "";
    $this->{inclusions} = ".*";
    $this->{exclusions} = "";
    $this->{override} = 0;
    $this->{toWeb} = "";
    $this->{addAfter} = "";
    $this->{addBefore} = "";
    $this->{parentTopic} = "";
    
    # informations for log
    $this->{logTmpl} = "advancedmanagecontriblog.tmpl";
    $this->{logTopic} = "AdvancedManageContribLog";
    $this->{logWeb} = "TWiki";
    $this->{logAffected} = "";
    $this->{logNotOverriden} = "";
    $this->{logExcluded} = "";
    $this->{logError} = "";
    
    &TWiki::Func::writeDebug( "AdvancedManageContrib: started" ) if $mydebug;
    
    
    # get all datas from form
    if ( defined($query->param('fromWeb')) && defined($query->param('toWeb')) ) {
      if ( &TWiki::Func::webExists($query->param('fromWeb')) && &TWiki::Func::webExists($query->param('toWeb'))) {
        $this->{fromWeb} = $query->param('fromWeb');
        $this->{toWeb} = $query->param('toWeb');
      } else {
        &TWiki::Func::writeDebug( "AdvancedManageContrib: fromWeb or toWeb are not defined, stop script" ) if $mydebug; 
        die "Unknown Web - Please notify your TWiki administrator";
      }
    } else {
     &TWiki::Func::writeDebug( "AdvancedManageContrib: fromWeb or toWeb are not defined, stop script" ) if $mydebug; 
     die "Webs not definied - Please notify your TWiki administrator";
    }
    
    if ( defined($query->param('inclusions')) ) {
      my $v = $query->param('inclusions');
      $v =~ s/([*?])/.$1/g;
      $v =~ s/,/|/g;
      $this->{inclusions} = $v;
    }
    
    if ( defined($query->param('exclusions')) ) {
      my $v = $query->param('exclusions');
      $v =~ s/([*?])/.$1/g;
      $v =~ s/,/|/g;
      $this->{exclusions} = $v;
    }
    
    if ( defined($query->param('override')) ) {
      $this->{override} = 1;
    }
    
    if ( defined($query->param('addition')) ) {
      if ( $query->param('addition') =~ /^(\w*)\*(\w*)$/ ) {
        $this->{addBefore} = $1;
        $this->{addAfter} = $2;
      }
    }
    
    if ( defined($query->param('parentTopic')) ) {
      $this->{parentTopic} = $query->param('parentTopic');
    }
    
    
    # get all topics to move
    my @topics = TWiki::Func::getTopicList($this->{fromWeb});
    
    foreach my $topic (@topics) {
      # analyze topic name, maybe the entry is like <web>.<topic>
      my $getWeb;
      if ( $topic =~ /^([A-z0-9]+)\.([A-z0-9_]+)/ ) {
        $this->{getWeb} = $1;
        $this->{getTopic} = $2;
        $this->{isChild} = 1;
      } else {
        $this->{getWeb} = $this->{fromWeb};
        $this->{getTopic} = $topic;
        $this->{isChild} = 0;
      }
      
      try {
        my $dispo = '';
        if( $this->{inclusions} && $this->{getTopic} !~ /^($this->{inclusions})$/ && !$this->{isChild} ) {
          $dispo = 'not included';
        } elsif( $this->{exclusions} && $this->{getTopic} =~ /^($this->{exclusions})$/ && !$this->{isChild} ) {
          $dispo = 'excluded';
          $this->{logExcluded} .= "| [[".$query->param('fromWeb').".".$this->{getTopic}."][".$query->param('toWeb').".".$this->{getTopic}."]] | \n";
        } else {
          $dispo = 'affected';
          if ( defined($query->param('MoveTopics')) ) {
            # check if topic-children affected
            if ( defined($query->param('moveChildren')) ) {
              my $children = $this->_getChildren($query,$session);
              foreach my $child (@$children) {
                push(@topics, $child);
                &TWiki::Func::writeDebug( "AdvancedManageContrib: Child: $child" ) if $mydebug;
              }
            }
            
            # move topic to new location
            $this->_moveTopic($query,$session);
            &TWiki::Func::writeDebug( "AdvancedManageContrib: affected: Web: " . $this->{getWeb} . " and Topic: " . $this->{getTopic}  ) if $mydebug;
            
          } elsif ( defined($query->param('SetParent')) ) {
            $this->_setParent($query,$session,$store);
          }
        }
        #&TWiki::Func::writeDebug( "AdvancedManageContrib: ".$this->{getTopic}." - $dispo" ) if $mydebug;
      } catch Error::Simple with {
        my $e = shift;
        &TWiki::Func::writeDebug( "AdvancedManageContrib: ERROR: Script got problems with topic: ".$this->{getTopic}." - ".$e->{-text} ) if $mydebug;
        $this->{logError} .= "| ".$this->{getTopic}." | ".$e->{-text}." |\n";
      };
    }
    
    # update log-topic
    my ( $meta, $text ) = &TWiki::Func::readTopic( $this->{logWeb}, $this->{logTopic} );
    
    my $templates = $session->templates;
    my $tmpl = $templates->readTemplate($this->{logTmpl});
    
    # build affected topics table
    my $affected = $templates->expandTemplate('affectedTopics');
    my $aff = $this->{logAffected};
    $affected =~ s/%AFFECTED_LIST%/$aff/go;
    $tmpl .= $affected;
    
    # build not overriden warnings table
    if ( $this->{logNotOverriden} ) {
      my $overriden = $templates->expandTemplate('notOverridenTopics');
      my $overr = $this->{logNotOverriden};
      $overriden =~ s/%NOTOVERRIDEN_LIST%/$overr/go;
      $tmpl .= $overriden;
    }
    
    # build excluded topics table
    if ( $this->{logExcluded} ) {
      my $excluded = $templates->expandTemplate('excludedTopics');
      my $excl = $this->{logExcluded};
      $excluded =~ s/%EXCLUDED_LIST%/$excl/go;
      $tmpl .= $excluded;
    }
    
    # build error log table
    if ( $this->{logError} ) {
      my $errors = $templates->expandTemplate('errorLog');
      my $err = $this->{logError};
      $errors =~ s/%ERROR_LIST%/$err/go;
      $tmpl .= $errors;
    }
    
    
    # store and goto log topic
    $store->saveTopic( $session->{user}, $this->{logWeb}, $this->{logTopic}, $tmpl, $meta, { } ); 
    my $url = &TWiki::Func::getViewUrl( $this->{logWeb}, $this->{logTopic} );
    TWiki::Func::redirectCgiQuery( $this->{query}, $url ); 
}


sub _moveTopic {
  my ($this, $query, $session) = @_;
  
  # set new twiki name
  my $newTopic = $this->{addBefore} . $this->{getTopic} . $this->{addAfter};
  
  if ( !&TWiki::Func::topicExists($query->param('toWeb'), $newTopic) || $this->{override} ) {
    # get all affected topics
    my $refs = &TWiki::UI::Manage::getReferringTopics( $session, $this->{getWeb}, $this->{getTopic}, 1 );
    my $refs0 = &TWiki::UI::Manage::getReferringTopics( $session, $this->{getWeb}, $this->{getTopic}, 0 );
    my @arrRefs;
    
    foreach my $ref (sort keys %$refs0) {
      $refs->{$ref} = $ref;
    }
    
    # put them all in an array
    foreach my $entry ( sort keys %$refs ) {
      $arrRefs[++$#arrRefs] = $entry;
      &TWiki::Func::writeDebug( "AdvancedManageContrib: Affected topic: " . $entry ) if $mydebug;
    }
    
    # move web from $this->{fromWeb}.$this->{getTopic} to $this->{toWeb}.$newTopic
    &TWiki::UI::Manage::move($session,$this->{getWeb},$this->{getTopic},$this->{toWeb},$newTopic,undef,\@arrRefs);
    
    $this->{logAffected} .= "| !" . $this->{getWeb} . ".".$this->{getTopic}." | moved to [[".$query->param('toWeb'). ".$newTopic][". $query->param('toWeb')
      .".$newTopic]] |\n";
  } else {
    $this->{logNotOverriden} .= "| [[".$query->param('toWeb').".$newTopic][".$query->param('toWeb').".$newTopic]] | already exists |\n";
  }
}


sub _setParent {
  my ($this, $query, $session, $store) = @_;
  
  # get topic infos
  my ( $meta, $text ) = &TWiki::Func::readTopic( $this->{getWeb}, $this->{getTopic} );
  
  # set parent
  if ( $this->{getWeb} eq $query->param('toWeb') ) {
    $meta->put( 'TOPICPARENT', { 'name' => "".$this->{parentTopic}."" } );
  } else {
    $meta->put( 'TOPICPARENT', { 'name' => "".$this->{toWeb}.".".$this->{parentTopic}."" } );
  }
  
  # try to save topic with new parent
  try {
    $store->saveTopic( $session->{user}, $this->{getWeb}, $this->{getTopic}, $text, $meta, { } );
    $this->{logAffected} .= "| ".$this->{getWeb}.".".$this->{getTopic}." | set parent to ".$query->param('toWeb').".".$this->{parentTopic}." |\n";
  } catch Error::Simple with {
    &TWiki::Func::writeDebug( "AdvancedManageContrib: ERROR: Cannot save topic: ".$this->{getTopic}." - ".shift->{-text} ) if $mydebug;
    $this->{logError} .= "| ".$this->{getWeb}." | ".shift->{-text}." |\n";
  }; 
  
}

sub _getChildren {
  my ($this, $query, $session) = @_;
  my @children;
  my @webs;
  
  # check if only same web should be checked for children
  if ( defined($query->param('sameWeb')) ) {
    push(@webs, $this->{getWeb});
  } else {
    @webs = TWiki::Func::getListOfWebs( "user" );
  }
  
  # foreach web -> foreach topic: check parents
  foreach my $web (@webs) {
     my @webTopics = TWiki::Func::getTopicList( $web );
     
     foreach my $webTopic (@webTopics) {
       my ( $meta, $text ) = &TWiki::Func::readTopic( $web, $webTopic );
       
       # check if $this->{getTopic} is parent
       if (($web == $this->{getWeb} && $meta->getParent() eq $this->{getTopic}) || 
            $meta->getParent() eq $this->{getWeb}.".".$this->{getTopic} ) {
         
         push(@children, $web.".".$webTopic); 
       }
    }
  }
  
  return \@children;
}

1;
